@@ -69,7 +69,11 @@ static void
 set_tooltip(GladeXML *xml, GtkWidget *widget,
 	    const gchar *prop_name, const gchar *prop_value)
 {
-    gtk_tooltips_set_tip(xml->priv->tooltips, widget, prop_value, NULL);
+    if (GTK_IS_TOOL_ITEM (widget))
+	gtk_tool_item_set_tooltip (GTK_TOOL_ITEM (widget), xml->priv->tooltips,
+				   prop_value, NULL);
+    else
+	gtk_tooltips_set_tip(xml->priv->tooltips, widget, prop_value, NULL);
 }
 
 static void
@@ -103,6 +107,7 @@ pixmap_set_filename (GladeXML *xml, GtkWidget *w,
     char *file;
     GdkPixmap *pixmap = NULL;
     GdkBitmap *bitmap = NULL;
+    GdkColormap *cmap;
     
     file = glade_xml_relative_file (xml, value);
     pb = gdk_pixbuf_new_from_file (file, NULL);
@@ -111,7 +116,8 @@ pixmap_set_filename (GladeXML *xml, GtkWidget *w,
     if (!pb)
 	return;
 
-    gdk_pixbuf_render_pixmap_and_mask (pb, &pixmap, &bitmap, 127);
+    cmap = gtk_widget_get_colormap (w);
+    gdk_pixbuf_render_pixmap_and_mask_for_colormap (pb, cmap, &pixmap, &bitmap, 127);
     gtk_pixmap_set (GTK_PIXMAP (w), pixmap, bitmap);
 
     if (pixmap) g_object_unref (pixmap);
@@ -375,6 +381,14 @@ menu_item_set_use_stock (GladeXML *xml, GtkWidget *w,
 
 	    gtk_label_set_text(GTK_LABEL(child), stock_item.label);
 	    gtk_label_set_use_underline(GTK_LABEL(child), TRUE);
+
+	    if (stock_item.keyval)
+		gtk_widget_add_accelerator (w,
+					    "activate",
+					    glade_xml_ensure_accel(xml),
+					    stock_item.keyval,
+					    stock_item.modifier,
+					    GTK_ACCEL_VISIBLE);
 	} else {
 	    g_warning("could not look up stock id '%s'", stock_id);
 	}
@@ -407,7 +421,11 @@ static void
 entry_set_invisible_char (GladeXML *xml, GtkWidget *w,
 			  const gchar *name, const gchar *value)
 {
-    gtk_entry_set_invisible_char (GTK_ENTRY (w), value [0]);
+    gunichar c;
+
+    c = g_utf8_get_char_validated (value, strlen (value));
+    if (c > 0)
+        gtk_entry_set_invisible_char (GTK_ENTRY (w), c);
 }
 
 static void
@@ -418,6 +436,82 @@ button_set_response_id (GladeXML *xml, GtkWidget *w,
 		       GINT_TO_POINTER (INT (value)));
 }
 
+static void
+toggle_tool_button_set_active (GladeXML *xml, GtkWidget *w,
+			       const char *name, const char *value)
+{
+    gtk_toggle_tool_button_set_active (GTK_TOGGLE_TOOL_BUTTON (w),
+				       BOOL (value));
+}
+
+static void
+tool_button_set_icon (GladeXML *xml, GtkWidget *w,
+		      const char *name, const char *value)
+{
+    GdkPixbuf *pb;
+    GtkWidget *image;
+    char *file;
+
+    file = glade_xml_relative_file (xml, value);
+    pb = gdk_pixbuf_new_from_file (file, NULL);
+    g_free (file);
+
+    if (!pb) {
+	g_warning ("Couldn't find image file: %s", value);
+	return;
+    }
+
+    image = gtk_image_new_from_pixbuf (pb);
+    g_object_unref (pb);
+
+    gtk_widget_show (image);
+    gtk_tool_button_set_icon_widget (GTK_TOOL_BUTTON (w), image);
+}
+
+static void
+combo_box_set_items (GladeXML *xml, GtkWidget *w,
+		     const char *name, const char *value)
+{
+    GtkListStore *store;
+    GtkCellRenderer *cell;
+    gchar *items, *items_end, *item_start, *item_end;
+    GtkTreeIter iter;
+
+    /* If the "items" property is set, we create a simple model with just
+       one column of text. */
+    store = gtk_list_store_new (1, G_TYPE_STRING);
+    gtk_combo_box_set_model (GTK_COMBO_BOX (w), GTK_TREE_MODEL (store));
+
+    /* GtkComboBoxEntry creates the cell renderer itself, but we have to set
+       the column containing the text. */
+    if (GTK_IS_COMBO_BOX_ENTRY (w)) {
+	gtk_combo_box_entry_set_text_column (GTK_COMBO_BOX_ENTRY (w), 0);
+    } else {
+	cell = gtk_cell_renderer_text_new ();
+	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (w), cell, TRUE);
+	gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (w), cell,
+					"text", 0,
+					NULL);
+    }
+
+    /* Now add the items one at a time. */
+    items = g_strdup (value);
+    items_end = &items[strlen (items)];
+    item_start = items;
+    while (item_start < items_end) {
+	item_end = strchr (item_start, '\n');
+	if (item_end == NULL)
+	    item_end = items_end;
+	*item_end = '\0';
+
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter, 0, item_start, -1);
+
+	item_start = item_end + 1;
+    }
+    g_free (items);
+}
+
 static void
 menuitem_build_children(GladeXML *self, GtkWidget *w,
 			GladeWidgetInfo *info)
@@ -515,6 +609,41 @@ frame_build_children(GladeXML *self, GtkWidget *parent,
     g_object_unref(G_OBJECT(parent));
 }
 
+static void
+expander_build_children (GladeXML        *self,
+			 GtkWidget       *parent,
+			 GladeWidgetInfo *info)
+{
+    int i;
+
+    g_object_ref (parent);
+    for (i = 0; i < info->n_children; i++) {
+	GladeWidgetInfo *childinfo = info->children [i].child;
+	GtkWidget       *child;
+	gboolean         label_item = FALSE;
+	int              j;
+	
+	child = glade_xml_build_widget (self, childinfo);
+
+	/* there should really only be 2 children */
+	for (j = 0; j < info->children [i].n_properties; j++) {
+	    if (!strcmp (info->children [i].properties[j].name, "type")) {
+		const char *value = info->children [i].properties [j].value;
+
+		if (!strcmp (value, "label_item"))
+		    label_item = TRUE;
+		break;
+	    }
+	}
+
+	if (label_item)
+	    gtk_expander_set_label_widget (GTK_EXPANDER (parent), child);
+	else
+	    gtk_container_add (GTK_CONTAINER (parent), child);
+    }
+    g_object_unref (parent);
+}
+
 static void
 notebook_build_children(GladeXML *self, GtkWidget *parent,
 			GladeWidgetInfo *info)
@@ -680,6 +809,7 @@ toolbar_build_children (GladeXML *xml, GtkWidget *parent,
 	    !strcmp (childinfo->child->classname, "radio") ||
 	    !strcmp (childinfo->child->classname, "button")) {
 	    const char *label = NULL, *stock = NULL, *group_name = NULL;
+	    const char *tooltip = NULL;
 	    char *icon = NULL;
 	    gboolean use_stock = FALSE, active = FALSE, new_group = FALSE;
 	    gboolean use_underline = FALSE;
@@ -711,7 +841,7 @@ toolbar_build_children (GladeXML *xml, GtkWidget *parent,
 		} else if (!strcmp (name, "visible")) {
 		    /* ignore for now */
 		} else if (!strcmp (name, "tooltip")) {
-		    /* ignore for now */
+		    tooltip = value;
 		} else if (!strcmp (name, "use_underline")) {
 		    use_underline = BOOL (value);
 		} else if (!strcmp (name, "inconsistent")) {
@@ -725,16 +855,8 @@ toolbar_build_children (GladeXML *xml, GtkWidget *parent,
 	       here, partly because GTK+ doesn't have direct support for stock
 	       toggle & radio items. */
 	    if (use_stock) {
-		GtkStockItem item;
-
-		if (gtk_stock_lookup (label, &item)) {
-		    /* Set stock to label, so the icon is created below. */
-		    stock = label;
-		    label = item.label;
-
-		    /* Most stock items have mnemonic accelerators. */
-		    use_underline = TRUE;
-		}
+		stock = label;
+		label = NULL;
 	    }
 
 	    if (stock) {
@@ -744,6 +866,7 @@ toolbar_build_children (GladeXML *xml, GtkWidget *parent,
 		    g_warning ("Could not create stock item: %s", stock);
 	    } else if (icon) {
 		iconw = gtk_image_new_from_file (icon);
+		g_free (icon);
 	    }
 
 	    if (iconw)
@@ -757,24 +880,26 @@ toolbar_build_children (GladeXML *xml, GtkWidget *parent,
 		child = gtk_toolbar_append_element (
 		    GTK_TOOLBAR (parent),
 		    GTK_TOOLBAR_CHILD_TOGGLEBUTTON, NULL,
-		    label, NULL, NULL, iconw, NULL, NULL);
-		gtk_toggle_button_set_active(
-		    GTK_TOGGLE_BUTTON (child), active);
+		    label, tooltip, NULL, iconw, NULL, NULL);
+		gtk_toggle_button_set_active
+		    (GTK_TOGGLE_BUTTON (child), active);
 	    } else if (!strcmp (childinfo->child->classname, "radio")) {
 		child = gtk_toolbar_append_element (
 		    GTK_TOOLBAR (parent),
 		    GTK_TOOLBAR_CHILD_RADIOBUTTON, NULL,
-		    label, NULL, NULL, iconw, NULL, NULL);
+		    label, tooltip, NULL, iconw, NULL, NULL);
 
 		if (group_name) {
 		    g_object_set (G_OBJECT (child),
 				  "group", glade_xml_get_widget (xml, group_name),
 				  NULL);
 		}
+		gtk_toggle_button_set_active
+		    (GTK_TOGGLE_BUTTON (child), active);
 	    } else
 		child = gtk_toolbar_append_item (
 		    GTK_TOOLBAR (parent),
-		    label, NULL, NULL, iconw, NULL, NULL);
+		    label, tooltip, NULL, iconw, NULL, NULL);
 	    
 	    /* GTK+ doesn't support use_underline directly, so we have to hack
 	       it. */
@@ -788,7 +913,11 @@ toolbar_build_children (GladeXML *xml, GtkWidget *parent,
 	    glade_xml_set_common_params (xml, child, childinfo->child);
 	} else {
 	    child = glade_xml_build_widget (xml, childinfo->child);
-	    gtk_toolbar_append_widget (GTK_TOOLBAR (parent), child, NULL, NULL);
+
+	    if (GTK_IS_TOOL_ITEM (child))
+		gtk_toolbar_insert (GTK_TOOLBAR (parent), GTK_TOOL_ITEM (child), -1);
+	    else
+		gtk_toolbar_append_widget (GTK_TOOLBAR (parent), child, NULL, NULL);
 	}
     }
 }
@@ -992,6 +1121,15 @@ combo_find_internal_child(GladeXML *xml, GtkWidget *parent,
     return NULL;
 }
 
+static GtkWidget *
+combo_box_entry_find_internal_child(GladeXML *xml, GtkWidget *parent,
+				    const gchar *childname)
+{
+    if (!strcmp(childname, "entry"))
+	return gtk_bin_get_child(GTK_BIN(parent));
+    return NULL;
+}
+
 void
 _glade_init_gtk_widgets(void)
 {
@@ -1030,7 +1168,12 @@ _glade_init_gtk_widgets(void)
     glade_register_custom_prop (GTK_TYPE_LIST_ITEM, "label", list_item_set_label);
     glade_register_custom_prop (GTK_TYPE_BUTTON, "response_id", button_set_response_id);
     glade_register_custom_prop (GTK_TYPE_ENTRY, "invisible_char", entry_set_invisible_char);
+    glade_register_custom_prop (GTK_TYPE_TOGGLE_TOOL_BUTTON, "active", toggle_tool_button_set_active);
+    glade_register_custom_prop (GTK_TYPE_TOOL_BUTTON, "icon", tool_button_set_icon);
+    glade_register_custom_prop (GTK_TYPE_COMBO_BOX, "items", combo_box_set_items);
 
+    glade_register_widget (GTK_TYPE_ABOUT_DIALOG, NULL,
+			   NULL, NULL);
     glade_register_widget (GTK_TYPE_ACCEL_LABEL, glade_standard_build_widget,
 			   NULL, NULL);
     glade_register_widget (GTK_TYPE_ALIGNMENT, glade_standard_build_widget,
@@ -1049,12 +1192,18 @@ _glade_init_gtk_widgets(void)
 			   menuitem_build_children, NULL);
     glade_register_widget (GTK_TYPE_CLIST, glade_standard_build_widget,
 			   clist_build_children, NULL);
+    glade_register_widget (GTK_TYPE_COLOR_BUTTON, glade_standard_build_widget,
+			   NULL, NULL);
     glade_register_widget (GTK_TYPE_COLOR_SELECTION, glade_standard_build_widget,
 			   NULL, NULL);
     glade_register_widget (GTK_TYPE_COLOR_SELECTION_DIALOG, NULL,
 			   glade_standard_build_children, colorseldlg_find_internal_child);
     glade_register_widget (GTK_TYPE_COMBO, glade_standard_build_widget,
 			   glade_standard_build_children, combo_find_internal_child);
+    glade_register_widget (GTK_TYPE_COMBO_BOX, glade_standard_build_widget,
+			   NULL, NULL);
+    glade_register_widget (GTK_TYPE_COMBO_BOX_ENTRY, glade_standard_build_widget,
+			   glade_standard_build_children, combo_box_entry_find_internal_child);
     glade_register_widget (GTK_TYPE_CTREE, glade_standard_build_widget,
 			   clist_build_children, NULL);
     glade_register_widget (GTK_TYPE_CURVE, glade_standard_build_widget,
@@ -1067,10 +1216,20 @@ _glade_init_gtk_widgets(void)
 			   NULL, NULL);
     glade_register_widget (GTK_TYPE_EVENT_BOX, glade_standard_build_widget,
 			   glade_standard_build_children, NULL);
+    glade_register_widget (GTK_TYPE_EXPANDER, glade_standard_build_widget,
+			   expander_build_children, NULL);
+    glade_register_widget (GTK_TYPE_FILE_CHOOSER, glade_standard_build_widget,
+			   NULL, NULL);
+    glade_register_widget (GTK_TYPE_FILE_CHOOSER_DIALOG, glade_standard_build_widget,
+			   gtk_dialog_build_children, dialog_find_internal_child);
+    glade_register_widget (GTK_TYPE_FILE_CHOOSER_BUTTON, glade_standard_build_widget,
+			   NULL, NULL);
     glade_register_widget (GTK_TYPE_FILE_SELECTION, NULL,
 			   glade_standard_build_children, filesel_find_internal_child);
     glade_register_widget (GTK_TYPE_FIXED, glade_standard_build_widget,
 			   glade_standard_build_children, NULL);
+    glade_register_widget (GTK_TYPE_FONT_BUTTON, glade_standard_build_widget,
+			   NULL, NULL);
     glade_register_widget (GTK_TYPE_FONT_SELECTION, glade_standard_build_widget,
 			   NULL, NULL);
     glade_register_widget (GTK_TYPE_FONT_SELECTION_DIALOG, NULL,
@@ -1095,6 +1254,8 @@ _glade_init_gtk_widgets(void)
 			   NULL, NULL);
     glade_register_widget (GTK_TYPE_HSEPARATOR, glade_standard_build_widget,
 			   NULL, NULL);
+    glade_register_widget (GTK_TYPE_ICON_VIEW, glade_standard_build_widget,
+			   NULL, NULL);
     glade_register_widget (GTK_TYPE_IMAGE, glade_standard_build_widget,
 			   NULL, NULL);
     glade_register_widget (GTK_TYPE_IMAGE_MENU_ITEM, glade_standard_build_widget,
@@ -1115,6 +1276,8 @@ _glade_init_gtk_widgets(void)
 			   glade_standard_build_children, NULL);
     glade_register_widget (GTK_TYPE_MENU_ITEM, glade_standard_build_widget,
 			   menuitem_build_children, NULL);
+    glade_register_widget (GTK_TYPE_MENU_TOOL_BUTTON, glade_standard_build_widget,
+			   NULL, NULL);
     glade_register_widget (GTK_TYPE_MESSAGE_DIALOG, NULL,
 			   glade_standard_build_children, NULL);
     glade_register_widget (GTK_TYPE_NOTEBOOK, glade_standard_build_widget,
@@ -1137,11 +1300,15 @@ _glade_init_gtk_widgets(void)
 			   glade_standard_build_children, NULL);
     glade_register_widget (GTK_TYPE_RADIO_MENU_ITEM, glade_standard_build_widget,
 			   menuitem_build_children, NULL);
+    glade_register_widget (GTK_TYPE_RADIO_TOOL_BUTTON, glade_standard_build_widget,
+			   NULL, NULL);
     glade_register_widget (GTK_TYPE_SCROLLED_WINDOW, glade_standard_build_widget,
 			   glade_standard_build_children,
 			   scrolled_window_find_internal_child);
     glade_register_widget (GTK_TYPE_SEPARATOR_MENU_ITEM, glade_standard_build_widget,
 			   NULL, NULL);
+    glade_register_widget (GTK_TYPE_SEPARATOR_TOOL_ITEM, glade_standard_build_widget,
+			   NULL, NULL);
 #ifdef HAVE_GTK_PLUG
     glade_register_widget (GTK_TYPE_SOCKET, glade_standard_build_widget,
 			   NULL, NULL);
@@ -1162,8 +1329,14 @@ _glade_init_gtk_widgets(void)
 			   NULL, NULL);
     glade_register_widget (GTK_TYPE_TOGGLE_BUTTON, glade_standard_build_widget,
 			   glade_standard_build_children, NULL);
+    glade_register_widget (GTK_TYPE_TOGGLE_TOOL_BUTTON, glade_standard_build_widget,
+			   NULL, NULL);
     glade_register_widget (GTK_TYPE_TOOLBAR, glade_standard_build_widget,
 			   toolbar_build_children, NULL);
+    glade_register_widget (GTK_TYPE_TOOL_ITEM, glade_standard_build_widget,
+			   glade_standard_build_children, NULL);
+    glade_register_widget (GTK_TYPE_TOOL_BUTTON, glade_standard_build_widget,
+			   NULL, NULL);
     glade_register_widget (GTK_TYPE_TREE, glade_standard_build_widget,
 			   NULL, NULL);
     glade_register_widget (GTK_TYPE_TREE_VIEW, glade_standard_build_widget,
